/*
 * pdb.c
 * Copyright (C) Christophe Schmitz and Thomas Huber 2008 <schmitz@maths.uq.edu.au> <numbat.pcs@gmail.com>
 *
 *
 *     This file is part of Numbat.
 * 
 *     Numbat is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 * 
 *     Numbat is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with Numbat.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib/gstdio.h>

#include "pdb.h"
#include "common.h"
#include "racs.h"
#include "pcs.h"
#include "methyl.h"
#include "select_display.h"
#include "tensor.h"
#include "rdc.h"
#include "error_analysis.h"


//~ When the user click on Load PDB button, it display the input
//~ file box, and call the function to read the data inside the pdb
//~ file to the _pdb_str
void
load_PDB (void)
{
	GtkWidget *dialog;
	GtkFileFilter *filter1;
	GtkFileFilter *filter2;
	gint result;
	int result_pdb;

	result_pdb = -1;

	dialog =
		gtk_file_chooser_dialog_new
		("Choose a pdb file containing a structure or family of structures",
		 NULL, GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL,
		 GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

	gtk_window_set_icon_from_file ((GtkWindow *) dialog, NUMBAT_ICON_FILE,
								   NULL);
	gtk_window_set_transient_for ((GtkWindow *) dialog, (GtkWindow *) _window);

	filter1 = gtk_file_filter_new ();
	filter2 = gtk_file_filter_new ();
	gtk_file_filter_add_pattern (filter1, "*.pdb");
	gtk_file_filter_set_name (filter1, "*.pdb");
	gtk_file_filter_add_pattern (filter2, "*.*");
	gtk_file_filter_set_name (filter2, "*.*");
	gtk_file_chooser_add_filter ((GtkFileChooser *) dialog, filter1);
	gtk_file_chooser_add_filter ((GtkFileChooser *) dialog, filter2);

	result = gtk_dialog_run (GTK_DIALOG (dialog));
	switch (result)
	{
	case GTK_RESPONSE_ACCEPT:
	{
		char *filename;

		filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
		result_pdb = read_pdb_file (filename);
		g_free (filename);

	}
		break;
	default:
		break;
	}
	gtk_object_destroy ((GtkObject *) filter1);
	gtk_object_destroy ((GtkObject *) filter2);
	gtk_widget_destroy (dialog);
}

//~ read_pdb_file is responsable of opening, reading and closing the pdb file.
//~ it take as arg the absolut file name, open it, count the number of model
//~ (for NMR structures) create the structure to store pdb informations,
//~ populate it.
int
read_pdb_file (char *name)
{
	FILE *f_pdb;
	int n_pdb_model;
	int n_pdb_atom;
	int result;
	int i;
	pdb_structure *local_pdb_str;

//First we try to open the file
	f_pdb = g_fopen (name, "r");
	if (f_pdb == NULL)
	{
		box_error ("Error, can't open the file %s\nno pdb loaded", name);
		return (-1);
	}
//Second we count how many model are there in the pdb
	n_pdb_model = count_pdb_model (f_pdb);
	if (n_pdb_model <= 0)
	{
		fclose (f_pdb);
		box_error
			("Error, no structure found in the file %s\nno pdb loaded", name);
		return (-2);
	}
	fprintf(stderr, "Found %d model in pdb\n", n_pdb_model);
//Third we count how many atoms are there
	n_pdb_atom = count_pdb_atom (f_pdb);
	fprintf(stderr, "Found %d atom in pdb\n", n_pdb_atom);
	if (n_pdb_atom <= 0)
	{
		fclose (f_pdb);
		box_error ("Error, couldn't find any atoms in the \
	 	different structures %s\nno pdb loaded", name);
		return (-3);
	}
//Fourth we create the structure
	local_pdb_str = new_pdb_structure (n_pdb_model, n_pdb_atom);
	if (NULL == local_pdb_str)
	{
		fclose (f_pdb);
		box_error
			("Error, can't allocate memory while opening %s\nno pdb loaded",
			 name);
		return (-4);
	}
//Fifth populate the structures
	result =
		populate_pdb_structure (n_pdb_model, n_pdb_atom, f_pdb, local_pdb_str);
	if (result <= 0)
	{
		free_pdb_structure (local_pdb_str);
		fclose (f_pdb);
		box_error
			("Error, can't populate pdb_structure while opening %s\nno pdb loaded",
			 name);
		return (-5);
	}

	fclose (f_pdb);

	//COMMIT START HERE
	//IT IS ONLY REACHED IF _pdb_str is valid.
	strcpy (local_pdb_str->file_name, name);
	free_pdb_structure (_pdb_str);
	_pdb_str = local_pdb_str;
	free_array_select_display ();
	create_array_select_display ();
//  set_title_window (_pdb_str->file_name);
	set_name_label_pdb (_pdb_str->file_name);
	clear_CH3 ();
	for (i = 0; i < N_PCS_ARRAY; i++)
	{							//we delete all PCS structures and CH3 association
		free_array_PCS (_str_pcs[i]);
		_str_pcs[i] = NULL;		//not done in free
		free_array_PCS (_all_pcs[i]);
		_all_pcs[i] = NULL;		//not done in free
		clear_tensor_valid (i);
	}


	populate_all_pcs ();		//pcs without exp values!

	for (i = 0; i < N_PCS_ARRAY; i++)
	{
		display_pcs_list (i, -2);
	}


	update_PDB_selector ();

	for (i = 0; i < N_PCS_ARRAY; i++)
	{
		display_PDB_selector (i);
	}
	display_PDB_selector (MULTIPLE_PCS_N_BUTTON);
	new_pdb_lodaded_for_rdc ();
	new_pdb_lodaded_for_racs ();
	update_CH3_tree_after_new_pdb ();

	return (1);
}

//~ this function count inside the pdb file how many MODEL are there
//~ (usually just 1, but more than one for NMR structures!)
int
count_pdb_model (FILE * f_pdb)
{
	int n;
	int flag;

	n = 0;
	flag = 0;
	fseek (f_pdb, 0, SEEK_SET);
	while (fgets (line, MAX_CAR, f_pdb) != NULL)
	{
		if (strncmp (line, "MODEL", 5) == 0)
		{
			n++;
		}
		if (strncmp (line, "ATOM", 4) == 0)
		{
			flag = 1;
		}
	}
	if (flag == 0)
	{
		return (0);
	}
	else
	{
		if (n >= 1)
		{
			return (n);
		}
		else
		{
			return (1);
		}
	}
}

//~ this function count inside the pdb file how many ATOM are there
//~ in the first MODEL. It doesn't check that all model are consistent
//~ (will be checked when populating)
int
count_pdb_atom (FILE * f_pdb)
{
	int n;

	n = 0;
	fseek (f_pdb, 0, SEEK_SET);
	while (fgets (line, MAX_CAR, f_pdb) != NULL)
	{
		if (strncmp (line, "ATOM", 4) == 0)
		{
			n++;
		}
		if (strncmp (line, "ENDMDL", 6) == 0)
		{
			return (n);
		}
		if (strncmp (line, "TER", 3) == 0)
		{
			return (n);
		}
	}
	return (n);
}

//~ this function is responsible for creation 
//~ and initialisation of the pdb_str
pdb_structure *
new_pdb_structure (int n_pdb_model, int n_pdb_atom)
{
	int i;
	int j;
	pdb_structure *local_pdb_str;
	local_pdb_str = (pdb_structure *) calloc (sizeof (pdb_structure), 1);
	local_pdb_str->tab_atom =
		(atom_table *) calloc (sizeof (atom_table), n_pdb_atom);
	if (local_pdb_str->tab_atom == NULL)
	{
		return (NULL);
	}
	for (i = 0; i < n_pdb_atom; i++)
	{
		local_pdb_str->tab_atom[i].atom_coo = (atom_coordinate *)
			calloc (sizeof (atom_coordinate), n_pdb_model);
		if (local_pdb_str->tab_atom[i].atom_coo == NULL)
		{
			for (j = 0; j < i; j++)
			{

				if ((local_pdb_str->tab_atom[j].atom_coo) != NULL)
				{
					free ((local_pdb_str->tab_atom[j].atom_coo));
					local_pdb_str->tab_atom[j].atom_coo = NULL;
				}
			}
			if ((local_pdb_str->tab_atom) != NULL)
			{
				free ((local_pdb_str->tab_atom));
				local_pdb_str->tab_atom = NULL;
			}
			return (NULL);
		}
	}
	strcpy (local_pdb_str->file_name, "");
	local_pdb_str->n_model = n_pdb_model;
	local_pdb_str->n_atom = n_pdb_atom;
	local_pdb_str->n_residue = 0;
	local_pdb_str->n_residue_type = 0;
	local_pdb_str->n_atom_type = 0;
	for (i = 0; i < MAX_ATOM_TYPE; i++)
	{
		strncpy (local_pdb_str->tab_atom_type[i], "\0\0\0\0\0\0\0",
				 MAX_ATOM_NAME_LENGTH);
	}
	for (i = 0; i < MAX_RESIDUE_TYPE; i++)
	{
		strncpy (local_pdb_str->tab_atom_type[i], "\0\0\0\0\0\0\0",
				 MAX_RESIDUE_NAME_LENGTH);
	}
	return (local_pdb_str);
}

//~ this function delete totally the previous pdb structure
//~ and initialize correctlyall parameters
int
free_pdb_structure (pdb_structure * local_pdb_str)
{
	int i;
	if (local_pdb_str == NULL)
	{
		return (1);
	}
	for (i = 0; i < local_pdb_str->n_atom; i++)
	{
		if ((local_pdb_str->tab_atom[i].atom_coo) != NULL)
		{
			free ((local_pdb_str->tab_atom[i].atom_coo));
			local_pdb_str->tab_atom[i].atom_coo = NULL;
		}
	}

	if ((local_pdb_str->tab_atom) != NULL)
	{
		free ((local_pdb_str->tab_atom));
		local_pdb_str->tab_atom = NULL;
	}

	free (local_pdb_str);
	local_pdb_str = NULL;

	return (1);
}


//~ this function display some info on the box about the pdb file
int
print_info_pdb_structure (void)
{
	char message[5000];
	if (_pdb_str != NULL)
	{
		sprintf (message, "\
		number of MODEL: %d\n\
		number of ATOM: %d\n\
		number of residue: %d\n\
		number of atom type: %d\n\
		number of residue type: %d\n\
		", _pdb_str->n_model, _pdb_str->n_atom, _pdb_str->n_residue, _pdb_str->n_atom_type, _pdb_str->n_residue_type);
		box_message (message);
	}
	return (1);
}

//~ This function is responsible for populating the pdb_structure.
//~ If there is inconsistency with the different strucutre, only 
//~ the first one will be kept
int
populate_pdb_structure (int n_pdb_model, int n_pdb_atom, FILE * f_pdb,
						pdb_structure * local_pdb_str)
{
	int i, j;					//i for model, j for atom
	int flag;
	int flag2;
	char string[50];
	int digit;
	int result;

	if (local_pdb_str == NULL)
	{
		return (-1);
	}
	local_pdb_str->n_model = n_pdb_model;
	local_pdb_str->n_atom = n_pdb_atom;
	local_pdb_str->n_residue = 0;
	local_pdb_str->n_atom_type = 0;
	local_pdb_str->n_residue_type = 0;

	fseek (f_pdb, 0, SEEK_SET);
	i = j = 0;
	flag = 0;
	while (fgets (line, MAX_CAR, f_pdb) != NULL)
	{
		if (strncmp (line, "MODEL", 5) == 0)
		{
			flag = 1;
			break;
		}
	}

	switch (flag)
	{
	case 1:
	{
		fseek (f_pdb, 0, SEEK_SET);
		for (i = 0; i < n_pdb_model; i++)
		{
			flag2 = 0;
			while (fgets (line, MAX_CAR, f_pdb) != NULL)
			{
				if (strncmp (line, "MODEL", 5) == 0)
				{
					flag2 = 1;
					break;
				}
			}
			if (flag2 == 0)
			{
				box_error ("Error, can't find enough MODEL\nno pdb loaded",
						   NULL);
				return (-1);
			}
			//we found a MODEL, is it the correct number?
			sscanf (line, "%s%d", string, &digit);
			if (digit != (i + 1))
			{
				box_error
					("Error, MODEL number not sequential\nno pdb loaded", NULL);
				return (-1);
			}
			//we found the correct MODEL, let's read it (similar to case '0' now)
			flag2 = 0;
			for (j = 0; j < n_pdb_atom; j++)
			{
				while (fgets (line, MAX_CAR, f_pdb) != NULL)
				{
					if (strncmp (line, "ATOM", 4) == 0)
					{
						flag2 = 1;
						break;
					}
				}
				if (flag2 == 0)
				{
					box_error ("Error, can't find enough atom\nno pdb loaded",
							   NULL);
					return (-1);
				}
				result = store (line, j, i, local_pdb_str);
				if (result < 0)
				{
					return (-1);
				}
			}

		}
		break;
	}
	case 0:
	{
		fseek (f_pdb, 0, SEEK_SET);
		flag2 = 0;
		for (j = 0; j < n_pdb_atom; j++)
		{
			while (fgets (line, MAX_CAR, f_pdb) != NULL)
			{
				if (strncmp (line, "ATOM", 4) == 0)
				{
					flag2 = 1;
					break;
				}
			}
			if (flag2 == 0)
			{
				box_error ("Error, can't find enough atom\nno pdb loaded",
						   NULL);
				return (-1);
			}
			result = store (line, j, 0, local_pdb_str);
			if (result < 0)
			{
				return (-1);
			}
		}
		break;
	}
	default:
	{
		break;
	}
	}

	result = populate_type (local_pdb_str);
	if (result < 0)
	{
		return (-1);
	}
	return (1);
}


//~ This function is responsible of the population of the 'type'
//~ information. Has to be done once the whole structure is already filled 
int
populate_type (pdb_structure * local_pdb_str)
{
	int i;						//loop over the atom number
	int j;						//loop over the atom_name
	int k;						//loop over the residue_name    
	int last_j;					//indice on the last atom_name not yet filled
	int last_k;					//indice on the last residue name not yet filled
	int last_residue_num;		//the last residue visited
	int current_residue_num;	//the current residue
	int number_of_residue;
	char at_name[MAX_ATOM_NAME_LENGTH];
	char re_name[MAX_RESIDUE_NAME_LENGTH];
	int flag;
	last_j = last_k = 0;
	number_of_residue = 0;
	last_residue_num = -9999;
	for (i = 0; i < local_pdb_str->n_atom; i++)
	{
		strcpy (at_name, local_pdb_str->tab_atom[i].atom_name);
		strcpy (re_name, local_pdb_str->tab_atom[i].residue_name);
		//first we deal with atom name
		flag = 0;
		for (j = 0; j < last_j; j++)
		{
			if (strcmp (at_name, local_pdb_str->tab_atom_type[j]) == 0)
			{
				flag = 1;
				break;
			}
		}
		if (flag == 0)
		{
			if (last_j == MAX_ATOM_TYPE)
			{
				return (-1);
			}
			strcpy (local_pdb_str->tab_atom_type[last_j], at_name);
			last_j++;
		}
		//second we deal with residue name
		flag = 0;
		for (k = 0; k < last_k; k++)
		{
			if (strcmp (re_name, local_pdb_str->tab_residue_type[k]) == 0)
			{
				flag = 1;
				break;
			}
		}
		if (flag == 0)
		{
			if (last_k == MAX_RESIDUE_TYPE)
			{
				return (-1);
			}
			strcpy (local_pdb_str->tab_residue_type[last_k], re_name);
			last_k++;
		}

		//third we count how many different residue are there
		current_residue_num = local_pdb_str->tab_atom[i].residue_num;
		if (current_residue_num != last_residue_num)
		{
			number_of_residue++;
			last_residue_num = current_residue_num;
		}
	}

	//we add the dummy atom!   
	strcpy (local_pdb_str->tab_atom_type[last_j], DUMMY_NAME);
	last_j++;

	local_pdb_str->n_residue = number_of_residue;
	local_pdb_str->n_atom_type = last_j;
	local_pdb_str->n_residue_type = last_k;
	return (1);
}

int
print_full_pdb_structure (pdb_structure * local_pdb_str)
{
	int i, j;
	for (i = 0; i < local_pdb_str->n_model; i++)
	{
		fprintf (stdout, "MODEL: %d\n", i + 1);
		for (j = 0; j < local_pdb_str->n_atom; j++)
		{
			fprintf (stdout, "%d\t%s\t%s\t%d\t%.3lf\t%.3lf\t%.3lf\n",
					 local_pdb_str->tab_atom[j].atom_num,
					 local_pdb_str->tab_atom[j].atom_name,
					 local_pdb_str->tab_atom[j].residue_name,
					 local_pdb_str->tab_atom[j].residue_num,
					 local_pdb_str->tab_atom[j].atom_coo[i].x,
					 local_pdb_str->tab_atom[j].atom_coo[i].y,
					 local_pdb_str->tab_atom[j].atom_coo[i].z);
		}
	}

	for (i = 0; i < local_pdb_str->n_residue_type; i++)
	{
		fprintf (stdout, "%s\n", local_pdb_str->tab_residue_type[i]);
	}
	fprintf (stdout, "\n");
	for (i = 0; i < local_pdb_str->n_atom_type; i++)
	{
		fprintf (stdout, "%s\n", local_pdb_str->tab_atom_type[i]);
	}
	return (1);
}

//~ store actually read a line of the pdb, and extract the informations
//~ that will be stored in the _pdb_str, at the correct location
//~ given by atom_num and model_num
int
store (char *line, int atom_num, int model_num, pdb_structure * local_pdb_str)
{
	int n;
	int at_num;
	int re_num;
	char at_name[MAX_ATOM_NAME_LENGTH];
	char re_name[MAX_RESIDUE_NAME_LENGTH];
	char alternate_position_id;	
	double x, y, z;
	n = 0;

	line[11] = '\0';
	line[20] = '\0';
	n = n + sscanf (line + 6, "%d", &at_num);
	n = n + sscanf (line + 17, "%s", re_name);
	alternate_position_id = line[16];	
	line[17] = '\0';
	n = n + sscanf (line + 12, "%s", at_name);
	n = n + sscanf (line + 22, "%d", &re_num);
	n = n + sscanf (line + 30, "%lf%lf%lf", &x, &y, &z);
	if (n != 7)
	{
		box_error ("The pdb file doesn't follow the specification\n\
                           check the number of space between column\nno pdb loaded", NULL);
		return (-1);
	}
	local_pdb_str->tab_atom[atom_num].atom_coo[model_num].x = x;
	local_pdb_str->tab_atom[atom_num].atom_coo[model_num].y = y;
	local_pdb_str->tab_atom[atom_num].atom_coo[model_num].z = z;
	if (model_num == 0)
	{
		local_pdb_str->tab_atom[atom_num].atom_num = at_num;
		strcpy (local_pdb_str->tab_atom[atom_num].atom_name, at_name);
		strcpy (local_pdb_str->tab_atom[atom_num].residue_name, re_name);
		local_pdb_str->tab_atom[atom_num].residue_num = re_num;
	}
	else
	{
		//because we want to allow consecutive atom numbering within different MODEL:
		at_num = local_pdb_str->tab_atom[atom_num].atom_num;
		if ((local_pdb_str->tab_atom[atom_num].atom_num != at_num) ||
			(local_pdb_str->tab_atom[atom_num].residue_num != re_num) ||
			(strcmp (local_pdb_str->tab_atom[atom_num].atom_name, at_name) != 0)
			|| (strcmp (local_pdb_str->tab_atom[atom_num].residue_name, re_name)
				!= 0))
		{
			box_error ("The different MODEl don't seem consistent\n\
                           no pdb loaded", NULL);
			fprintf (stderr,
					 "at_num: %d re_num: %d at_name: %s re_name:%s from model %d\n",
					 at_num, re_num, at_name, re_name, model_num + 1);
			fprintf (stderr, "should be equal to\n");
			fprintf (stderr,
					 "at_num: %d re_num: %d at_name: %s re_name:%s from model %d\n",
					 local_pdb_str->tab_atom[atom_num].atom_num,
					 local_pdb_str->tab_atom[atom_num].residue_num,
					 local_pdb_str->tab_atom[atom_num].atom_name,
					 local_pdb_str->tab_atom[atom_num].residue_name, 1);

			return (-1);
		}
	}
	return (1);
}

//~ initialize correctly the pdb_str. called only once at the begining.
//~ to delete (and repopulate after the pdb_str), use
//~ delete_pdb_structure ()
int
init_pdb_structure (void)
{
	_pdb_str = NULL;
	return (1);
}

int
find_residue_type_in_pdb (char *name)
{
	int i;
	for (i = 0; i < _pdb_str->n_residue_type; i++)
	{
		if (strcmp (_pdb_str->tab_residue_type[i], name) == 0)
		{
			return (i);
		}
	}
	return (-1);
}

int
find_atom_type_in_pdb (char *name)
{
	int i;
	for (i = 0; i < _pdb_str->n_atom_type; i++)
	{
		if (strcmp (_pdb_str->tab_atom_type[i], name) == 0)
		{
			return (i);
		}
	}
	return (-1);
}
